
from typing import Any, List, Optional, Tuple, TYPE_CHECKING

from ...storage.models.atom import Atom, AtomKind, ZERO32, bytes_list_to_atoms
from .accounts import Accounts
from .receipt import Receipt
from .transaction import apply_transaction

if TYPE_CHECKING:
    from ...storage.models.trie import Trie
    from .transaction import Transaction
    from .receipt import Receipt

def _int_to_be_bytes(n: Optional[int]) -> bytes:
    if n is None:
        return b""
    n = int(n)
    if n == 0:
        return b"\x00"
    size = (n.bit_length() + 7) // 8
    return n.to_bytes(size, "big")


def _be_bytes_to_int(b: Optional[bytes]) -> int:
    if not b:
        return 0
    return int.from_bytes(b, "big")


class Block:
    """Validation Block representation using Atom storage.

    Top-level encoding:
      block_id = type_atom.object_id()
      chain: type_atom --next--> version_atom --next--> signature_atom --next--> body_list_atom --next--> ZERO32
      where: type_atom        = Atom(kind=AtomKind.SYMBOL, data=b"block")
             version_atom     = Atom(kind=AtomKind.BYTES,  data=b"\x01")
             signature_atom   = Atom(kind=AtomKind.BYTES, data=<signature-bytes>)
             body_list_atom   = Atom(kind=AtomKind.LIST,  data=<body_head_id>)

    Details order in body_list:
      0: chain                               (byte)
      1: previous_block_hash                 (bytes)
      2: number                              (int -> big-endian bytes)
      3: timestamp                           (int -> big-endian bytes)
      4: accounts_hash                       (bytes)
      5: transactions_total_fees             (int -> big-endian bytes)
      6: transactions_hash                   (bytes)
      7: receipts_hash                       (bytes)
      8: delay_difficulty                    (int -> big-endian bytes)
      9: validator_public_key_bytes         (bytes)
      10: nonce                              (int -> big-endian bytes)

    Notes:
      - "body tree" is represented here by the body_list id (self.body_hash), not
        embedded again as a field to avoid circular references.
      - "signature" is a field on the class but is not required for validation
        navigation; include it in the instance but it is not encoded in atoms
        unless explicitly provided via details extension in the future.
    """

    # essential identifiers
    version: int
    atom_hash: Optional[bytes]
    chain_id: int
    previous_block_hash: bytes
    previous_block: Optional["Block"]

    # block details
    number: int
    timestamp: Optional[int]
    accounts_hash: Optional[bytes]
    transactions_total_fees: Optional[int]
    transactions_hash: Optional[bytes]
    receipts_hash: Optional[bytes]
    delay_difficulty: Optional[int]
    validator_public_key_bytes: Optional[bytes]
    nonce: Optional[int]

    # additional
    body_hash: Optional[bytes]
    signature: Optional[bytes]

    # structures
    accounts: Optional["Trie"]
    transactions: Optional[List["Transaction"]]
    receipts: Optional[List["Receipt"]]
    
    def __init__(
        self,
        *,
        chain_id: int,
        previous_block_hash: bytes,
        previous_block: Optional["Block"],
        number: int,
        timestamp: Optional[int],
        accounts_hash: Optional[bytes],
        transactions_total_fees: Optional[int],
        transactions_hash: Optional[bytes],
        receipts_hash: Optional[bytes],
        delay_difficulty: Optional[int],
        validator_public_key_bytes: Optional[bytes],
        version: int = 1,
        nonce: Optional[int] = None,
        signature: Optional[bytes] = None,
        atom_hash: Optional[bytes] = None,
        body_hash: Optional[bytes] = None,
        accounts: Optional["Trie"] = None,
        transactions: Optional[List["Transaction"]] = None,
        receipts: Optional[List["Receipt"]] = None,
    ) -> None:
        self.version = int(version)
        self.atom_hash = atom_hash
        self.chain_id = chain_id
        self.previous_block_hash = previous_block_hash
        self.previous_block = previous_block
        self.number = number
        self.timestamp = timestamp
        self.accounts_hash = accounts_hash
        self.transactions_total_fees = transactions_total_fees
        self.transactions_hash = transactions_hash
        self.receipts_hash = receipts_hash
        self.delay_difficulty = delay_difficulty
        self.validator_public_key_bytes = (
            bytes(validator_public_key_bytes) if validator_public_key_bytes else None
        )
        self.nonce = nonce
        self.body_hash = body_hash
        self.signature = signature
        self.accounts = accounts
        self.transactions = transactions
        self.receipts = receipts

    def to_atom(self) -> Tuple[bytes, List[Atom]]:
        # Build body details as direct byte atoms, in defined order
        detail_payloads: List[bytes] = []
        block_atoms: List[Atom] = []

        def _emit(detail_bytes: bytes) -> None:
            detail_payloads.append(detail_bytes)

        # 0: chain
        _emit(_int_to_be_bytes(self.chain_id))
        # 1: previous_block_hash
        _emit(self.previous_block_hash)
        # 2: number
        _emit(_int_to_be_bytes(self.number))
        # 3: timestamp
        _emit(_int_to_be_bytes(self.timestamp))
        # 4: accounts_hash
        _emit(self.accounts_hash or b"")
        # 5: transactions_total_fees
        _emit(_int_to_be_bytes(self.transactions_total_fees))
        # 6: transactions_hash
        _emit(self.transactions_hash or b"")
        # 7: receipts_hash
        _emit(self.receipts_hash or b"")
        # 8: delay_difficulty
        _emit(_int_to_be_bytes(self.delay_difficulty))
        # 9: validator_public_key_bytes
        _emit(self.validator_public_key_bytes or b"")
        # 10: nonce
        _emit(_int_to_be_bytes(self.nonce))

        # Build body list chain directly from detail atoms
        body_head = ZERO32
        detail_atoms: List[Atom] = []
        for payload in reversed(detail_payloads):
            atom = Atom(data=payload, next_id=body_head, kind=AtomKind.BYTES)
            detail_atoms.append(atom)
            body_head = atom.object_id()
        detail_atoms.reverse()

        block_atoms.extend(detail_atoms)

        body_list_atom = Atom(data=body_head, kind=AtomKind.LIST)
        self.body_hash = body_list_atom.object_id()

        # Signature atom links to body list atom; type atom links to signature atom
        sig_atom = Atom(
            data=bytes(self.signature or b""),
            next_id=self.body_hash,
            kind=AtomKind.BYTES,
        )
        version_atom = Atom(
            data=_int_to_be_bytes(self.version),
            next_id=sig_atom.object_id(),
            kind=AtomKind.BYTES,
        )
        type_atom = Atom(
            data=b"block",
            next_id=version_atom.object_id(),
            kind=AtomKind.SYMBOL,
        )

        block_atoms.append(body_list_atom)
        block_atoms.append(sig_atom)
        block_atoms.append(version_atom)
        block_atoms.append(type_atom)

        self.atom_hash = type_atom.object_id()
        return self.atom_hash, block_atoms

    @classmethod
    def from_atom(cls, node: Any, block_id: bytes) -> "Block":

        block_header = node.get_atom_list(block_id)
        if block_header is None or len(block_header) != 4:
            raise ValueError("malformed block atom chain")
        type_atom, version_atom, sig_atom, body_list_atom = block_header

        if type_atom.kind is not AtomKind.SYMBOL or type_atom.data != b"block":
            raise ValueError("not a block (type atom payload)")
        if version_atom.kind is not AtomKind.BYTES:
            raise ValueError("malformed block (version atom kind)")
        version = _be_bytes_to_int(version_atom.data)
        if version != 1:
            raise ValueError("unsupported block version")
        if sig_atom.kind is not AtomKind.BYTES:
            raise ValueError("malformed block (signature atom kind)")
        if body_list_atom.kind is not AtomKind.LIST:
            raise ValueError("malformed block (body list atom kind)")
        if body_list_atom.next_id != ZERO32:
            raise ValueError("malformed block (body list tail)")

        detail_atoms = node.get_atom_list(body_list_atom.data)
        if detail_atoms is None:
            raise ValueError("missing block body list nodes")

        if len(detail_atoms) != 11:
            raise ValueError("block body must contain exactly 11 detail entries")

        detail_values: List[bytes] = []
        for detail_atom in detail_atoms:
            if detail_atom.kind is not AtomKind.BYTES:
                raise ValueError("block body detail atoms must be bytes")
            detail_values.append(detail_atom.data)

        (
            chain_bytes,
            prev_bytes,
            number_bytes,
            timestamp_bytes,
            accounts_bytes,
            fees_bytes,
            transactions_bytes,
            receipts_bytes,
            delay_diff_bytes,
            validator_bytes,
            nonce_bytes,
        ) = detail_values

        return cls(
            version=version,
            chain_id=_be_bytes_to_int(chain_bytes),
            previous_block_hash=prev_bytes or ZERO32,
            previous_block=None,
            number=_be_bytes_to_int(number_bytes),
            timestamp=_be_bytes_to_int(timestamp_bytes),
            accounts_hash=accounts_bytes or None,
            transactions_total_fees=_be_bytes_to_int(fees_bytes),
            transactions_hash=transactions_bytes or None,
            receipts_hash=receipts_bytes or None,
            delay_difficulty=_be_bytes_to_int(delay_diff_bytes),
            validator_public_key_bytes=validator_bytes or None,
            nonce=_be_bytes_to_int(nonce_bytes),
            signature=sig_atom.data if sig_atom is not None else None,
            atom_hash=block_id,
            body_hash=body_list_atom.object_id(),
        )

    def verify(self, node: Any) -> bool:
        """Verify receipts, transactions, and accounts hashes for this block."""
        if node is None:
            raise ValueError("node required for block verification")

        logger = getattr(node, "logger", None)

        def _hex(value: Optional[bytes]) -> str:
            if isinstance(value, (bytes, bytearray)):
                return value.hex()
            return str(value)

        def _log_debug(message: str, *args: object) -> None:
            if logger:
                logger.debug(message, *args)

        def _log_warning(message: str, *args: object) -> None:
            if logger:
                logger.warning(message, *args)

        _log_debug("Block verify start block=%s", _hex(self.atom_hash))

        if self.transactions_hash is None:
            _log_warning("Block verify missing transactions_hash block=%s", _hex(self.atom_hash))
            return False
        if self.receipts_hash is None:
            _log_warning("Block verify missing receipts_hash block=%s", _hex(self.atom_hash))
            return False
        if self.accounts_hash is None:
            _log_warning("Block verify missing accounts_hash block=%s", _hex(self.atom_hash))
            return False

        def _load_hash_list(head: bytes) -> Optional[List[bytes]]:
            if head == ZERO32:
                return []
            atoms = node.get_atom_list(head)
            if atoms is None:
                _log_warning("Block verify missing list atoms head=%s block=%s", _hex(head), _hex(self.atom_hash))
                return None
            for atom in atoms:
                if atom.kind is not AtomKind.BYTES:
                    _log_warning("Block verify list atom kind mismatch head=%s block=%s", _hex(head), _hex(self.atom_hash))
                    return None
            return [bytes(atom.data) for atom in atoms]

        prev_hash = self.previous_block_hash or ZERO32
        if prev_hash == ZERO32:
            if self.transactions_hash != ZERO32:
                _log_warning("Block verify genesis tx hash mismatch block=%s", _hex(self.atom_hash))
                return False
            if self.receipts_hash != ZERO32:
                _log_warning("Block verify genesis receipts hash mismatch block=%s", _hex(self.atom_hash))
                return False
            if self.transactions_total_fees not in (0, None):
                _log_warning("Block verify genesis fees mismatch block=%s", _hex(self.atom_hash))
                return False
            _log_debug("Block verify genesis passed block=%s", _hex(self.atom_hash))
            return True

        try:
            prev_block = self.previous_block or Block.from_atom(node, prev_hash)
        except Exception:
            _log_warning("Block verify failed loading parent block=%s prev=%s", _hex(self.atom_hash), _hex(prev_hash))
            return False

        prev_accounts_hash = getattr(prev_block, "accounts_hash", None)
        if not prev_accounts_hash:
            _log_warning("Block verify missing parent accounts hash block=%s", _hex(self.atom_hash))
            return False

        tx_hashes = _load_hash_list(self.transactions_hash)
        if tx_hashes is None:
            _log_warning("Block verify failed loading tx list block=%s", _hex(self.atom_hash))
            return False

        expected_tx_head, _ = bytes_list_to_atoms(tx_hashes)
        if expected_tx_head != (self.transactions_hash or ZERO32):
            _log_warning(
                "Block verify tx head mismatch block=%s expected=%s actual=%s",
                _hex(self.atom_hash),
                _hex(expected_tx_head),
                _hex(self.transactions_hash),
            )
            return False

        accounts_snapshot = Accounts(root_hash=prev_accounts_hash)
        work_block = type("_WorkBlock", (), {})()
        work_block.chain_id = self.chain_id
        work_block.accounts = accounts_snapshot
        work_block.transactions = []
        work_block.receipts = []

        total_fees = 0
        for tx_hash in tx_hashes:
            try:
                total_fees += apply_transaction(node, work_block, tx_hash)
            except Exception:
                _log_warning(
                    "Block verify failed applying tx=%s block=%s",
                    _hex(tx_hash),
                    _hex(self.atom_hash),
                )
                return False

        if self.transactions_total_fees is None:
            _log_warning("Block verify missing total fees block=%s", _hex(self.atom_hash))
            return False
        if int(self.transactions_total_fees) != int(total_fees):
            _log_warning(
                "Block verify fees mismatch block=%s expected=%s actual=%s",
                _hex(self.atom_hash),
                total_fees,
                self.transactions_total_fees,
            )
            return False

        applied_transactions = list(work_block.transactions or [])
        if len(applied_transactions) != len(tx_hashes):
            _log_warning(
                "Block verify tx count mismatch block=%s expected=%s actual=%s",
                _hex(self.atom_hash),
                len(tx_hashes),
                len(applied_transactions),
            )
            return False

        expected_receipts: List[Receipt] = list(work_block.receipts or [])
        if len(expected_receipts) != len(applied_transactions):
            _log_warning(
                "Block verify receipt count mismatch block=%s expected=%s actual=%s",
                _hex(self.atom_hash),
                len(applied_transactions),
                len(expected_receipts),
            )
            return False
        expected_receipt_ids: List[bytes] = []
        for receipt in expected_receipts:
            receipt_id, _ = receipt.to_atom()
            expected_receipt_ids.append(receipt_id)

        expected_receipts_head, _ = bytes_list_to_atoms(expected_receipt_ids)
        if expected_receipts_head != (self.receipts_hash or ZERO32):
            _log_warning(
                "Block verify receipts head mismatch block=%s expected=%s actual=%s",
                _hex(self.atom_hash),
                _hex(expected_receipts_head),
                _hex(self.receipts_hash),
            )
            return False

        stored_receipt_ids = _load_hash_list(self.receipts_hash)
        if stored_receipt_ids is None:
            _log_warning("Block verify failed loading receipts list block=%s", _hex(self.atom_hash))
            return False
        if stored_receipt_ids != expected_receipt_ids:
            _log_warning("Block verify receipts list mismatch block=%s", _hex(self.atom_hash))
            return False
        for expected, stored_id in zip(expected_receipts, stored_receipt_ids):
            try:
                stored = Receipt.from_atom(node, stored_id)
            except Exception:
                _log_warning(
                    "Block verify failed loading receipt=%s block=%s",
                    _hex(stored_id),
                    _hex(self.atom_hash),
                )
                return False
            if stored.transaction_hash != expected.transaction_hash:
                _log_warning(
                    "Block verify receipt tx mismatch receipt=%s block=%s",
                    _hex(stored_id),
                    _hex(self.atom_hash),
                )
                return False
            if stored.status != expected.status:
                _log_warning(
                    "Block verify receipt status mismatch receipt=%s block=%s",
                    _hex(stored_id),
                    _hex(self.atom_hash),
                )
                return False
            if stored.cost != expected.cost:
                _log_warning(
                    "Block verify receipt cost mismatch receipt=%s block=%s",
                    _hex(stored_id),
                    _hex(self.atom_hash),
                )
                return False
            if stored.logs_hash != expected.logs_hash:
                _log_warning(
                    "Block verify receipt logs hash mismatch receipt=%s block=%s",
                    _hex(stored_id),
                    _hex(self.atom_hash),
                )
                return False

        try:
            accounts_snapshot.update_trie(node)
        except Exception:
            _log_warning("Block verify failed updating accounts trie block=%s", _hex(self.atom_hash))
            return False
        if accounts_snapshot.root_hash != self.accounts_hash:
            _log_warning(
                "Block verify accounts hash mismatch block=%s expected=%s actual=%s",
                _hex(self.atom_hash),
                _hex(accounts_snapshot.root_hash),
                _hex(self.accounts_hash),
            )
            return False

        _log_debug("Block verify success block=%s", _hex(self.atom_hash))
        return True

    @staticmethod
    def _leading_zero_bits(buf: bytes) -> int:
        """Return the number of leading zero bits in the provided buffer."""
        zeros = 0
        for byte in buf:
            if byte == 0:
                zeros += 8
                continue
            zeros += 8 - int(byte).bit_length()
            break
        return zeros

    @staticmethod
    def calculate_delay_difficulty(
        *,
        previous_timestamp: Optional[int],
        current_timestamp: Optional[int],
        previous_difficulty: Optional[int],
        target_spacing: int = 2,
    ) -> int:
        """
        Adjust the delay difficulty with linear steps relative to block spacing.

        If blocks arrive too quickly (spacing <= 1), difficulty increases by one.
        If blocks are slower than the target spacing, difficulty decreases by one,
        and otherwise remains unchanged.
        """
        base_difficulty = max(1, int(previous_difficulty or 1))
        if previous_timestamp is None or current_timestamp is None:
            return base_difficulty

        spacing = max(0, int(current_timestamp) - int(previous_timestamp))
        if spacing <= 1:
            return base_difficulty + 1
        if spacing > target_spacing:
            return max(1, base_difficulty - 1)
        return base_difficulty

    def generate_nonce(
        self,
        *,
        difficulty: int,
    ) -> int:
        """
        Find a nonce that yields a block hash with the required leading zero bits.

        The search starts from the current nonce and iterates until the target
        difficulty is met.
        """
        target = max(1, int(difficulty))
        start = int(self.nonce or 0)
        nonce = start
        while True:
            self.nonce = nonce
            block_hash, _ = self.to_atom()
            leading_zeros = self._leading_zero_bits(block_hash)
            if leading_zeros >= target:
                self.atom_hash = block_hash
                return nonce
            nonce += 1



import logging
import socket
from enum import IntEnum
from typing import Optional, TYPE_CHECKING, Tuple

from .object_response import (
    ObjectResponse,
    ObjectResponseType,
    OBJECT_FOUND_ATOM_PAYLOAD,
    OBJECT_FOUND_LIST_PAYLOAD,
    encode_object_found_atom_payload,
    encode_object_found_list_payload,
)
from ..outgoing_queue import enqueue_outgoing
from ..models.message import Message, MessageTopic
from ..util import xor_distance
from ...storage.providers import provider_id_for_payload, provider_payload_for_id

if TYPE_CHECKING:
    from .. import Node
    from ..models.peer import Peer


class ObjectRequestType(IntEnum):
    OBJECT_GET = 0
    OBJECT_PUT = 1


class ObjectRequest:
    type: ObjectRequestType
    data: bytes
    atom_id: bytes
    payload_type: Optional[int]

    def __init__(
        self,
        type: ObjectRequestType,
        data: bytes = b"",
        atom_id: bytes = None,
        payload_type: Optional[int] = None,
    ):
        self.type = type
        self.data = data
        self.atom_id = atom_id
        self.payload_type = payload_type

    def to_bytes(self):
        if self.type == ObjectRequestType.OBJECT_PUT and self.payload_type is None:
            raise ValueError("OBJECT_PUT requires payload_type")
        if self.payload_type is not None:
            payload = bytes([self.payload_type]) + self.data
        else:
            payload = self.data
        return bytes([self.type.value]) + self.atom_id + payload

    @classmethod
    def from_bytes(cls, data: bytes) -> "ObjectRequest":
        # need at least 1 byte for type + 32 bytes for hash
        if len(data) < 1 + 32:
            raise ValueError(f"Too short for ObjectRequest ({len(data)} bytes)")

        type_val = data[0]
        try:
            req_type = ObjectRequestType(type_val)
        except ValueError:
            raise ValueError(f"Unknown ObjectRequestType: {type_val!r}")

        atom_id_bytes = data[1:33]
        payload = data[33:]
        if req_type == ObjectRequestType.OBJECT_GET:
            if payload:
                payload_type = payload[0]
                payload = payload[1:]
            else:
                payload_type = None
            return cls(req_type, payload, atom_id_bytes, payload_type=payload_type)
        if req_type == ObjectRequestType.OBJECT_PUT:
            if not payload:
                raise ValueError("OBJECT_PUT missing payload type")
            payload_type = payload[0]
            payload = payload[1:]
            return cls(req_type, payload, atom_id_bytes, payload_type=payload_type)
        return cls(req_type, payload, atom_id_bytes)


def encode_peer_contact_bytes(peer: "Peer") -> bytes:
    """Return a fixed-width peer contact payload (32-byte key + IPv4 + port)."""
    host, port = peer.address
    key_bytes = peer.public_key_bytes
    try:
        ip_bytes = socket.inet_aton(host)
    except OSError as exc:  # pragma: no cover - inet_aton raises for invalid hosts
        raise ValueError(f"invalid IPv4 address: {host}") from exc
    if not (0 <= port <= 0xFFFF):
        raise ValueError(f"port out of range (0-65535): {port}")
    port_bytes = int(port).to_bytes(2, "big", signed=False)
    return key_bytes + ip_bytes + port_bytes


def handle_object_request(node: "Node", peer: "Peer", message: Message) -> None:
    if message.content is None:
        node.logger.warning("OBJECT_REQUEST from %s missing content", peer.address)
        return

    try:
        object_request = ObjectRequest.from_bytes(message.content)
    except Exception as exc:
        node.logger.warning("Error decoding OBJECT_REQUEST from %s: %s", peer.address, exc)
        return

    match object_request.type:
        case ObjectRequestType.OBJECT_GET:
            atom_id = object_request.atom_id
            node.logger.debug("Handling OBJECT_GET for %s from %s", atom_id.hex(), peer.address)
            payload_type = object_request.payload_type
            if payload_type is None:
                payload_type = OBJECT_FOUND_ATOM_PAYLOAD

            if payload_type == OBJECT_FOUND_ATOM_PAYLOAD:
                local_atom = node.get_atom_from_local_storage(atom_id=atom_id)
                if local_atom is not None:
                    node.logger.debug("Object %s found locally; returning to %s", atom_id.hex(), peer.address)
                    resp = ObjectResponse(
                        type=ObjectResponseType.OBJECT_FOUND,
                        data=encode_object_found_atom_payload(local_atom),
                        atom_id=atom_id
                    )
                    obj_res_msg  = Message(
                        topic=MessageTopic.OBJECT_RESPONSE,
                        body=resp.to_bytes(),
                        sender=node.relay_public_key,
                    )
                    obj_res_msg.encrypt(peer.shared_key_bytes)
                    enqueue_outgoing(
                        node,
                        peer.address,
                        message=obj_res_msg,
                        difficulty=peer.difficulty,
                    )
                    return
            elif payload_type == OBJECT_FOUND_LIST_PAYLOAD:
                node.logger.debug(
                    "OBJECT_GET list request atom_id=%s from=%s",
                    atom_id.hex(),
                    peer.address,
                )
                local_atoms = node.get_atom_list_from_local_storage(root_hash=atom_id)
                if local_atoms is not None:
                    node.logger.debug("Object list %s found locally; returning to %s", atom_id.hex(), peer.address)
                    resp = ObjectResponse(
                        type=ObjectResponseType.OBJECT_FOUND,
                        data=encode_object_found_list_payload(local_atoms),
                        atom_id=atom_id
                    )
                    obj_res_msg  = Message(
                        topic=MessageTopic.OBJECT_RESPONSE,
                        body=resp.to_bytes(),
                        sender=node.relay_public_key,
                    )
                    obj_res_msg.encrypt(peer.shared_key_bytes)
                    enqueue_outgoing(
                        node,
                        peer.address,
                        message=obj_res_msg,
                        difficulty=peer.difficulty,
                    )
                    return
            else:
                node.logger.warning(
                    "Unknown OBJECT_GET payload type %s for %s",
                    payload_type,
                    atom_id.hex(),
                )

            if atom_id in node.storage_index:
                provider_id = node.storage_index[atom_id]
                provider_bytes = provider_payload_for_id(node, provider_id)
                if provider_bytes is not None:
                    node.logger.debug("Known provider for %s; informing %s", atom_id.hex(), peer.address)
                    resp = ObjectResponse(
                        type=ObjectResponseType.OBJECT_PROVIDER,
                        data=provider_bytes,
                        atom_id=atom_id
                    )
                    obj_res_msg = Message(
                        topic=MessageTopic.OBJECT_RESPONSE,
                        body=resp.to_bytes(),
                        sender=node.relay_public_key,
                    )
                    obj_res_msg.encrypt(peer.shared_key_bytes)
                    enqueue_outgoing(
                        node,
                        peer.address,
                        message=obj_res_msg,
                        difficulty=peer.difficulty,
                    )
                    return
                node.logger.warning(
                    "Unknown provider id %s for %s",
                    provider_id,
                    atom_id.hex(),
                )

            nearest_peer = node.peer_route.closest_peer_for_hash(atom_id)
            if nearest_peer:
                node.logger.debug("Forwarding requester %s to nearest peer for %s", peer.address, atom_id.hex())
                peer_info = encode_peer_contact_bytes(nearest_peer)
                resp = ObjectResponse(
                    type=ObjectResponseType.OBJECT_PROVIDER,
                    # type=ObjectResponseType.OBJECT_NEAREST_PEER,
                    data=peer_info,
                    atom_id=atom_id
                )
                obj_res_msg = Message(
                    topic=MessageTopic.OBJECT_RESPONSE,
                    body=resp.to_bytes(),
                    sender=node.relay_public_key,
                )
                obj_res_msg.encrypt(peer.shared_key_bytes)
                enqueue_outgoing(
                    node,
                    peer.address,
                    message=obj_res_msg,
                    difficulty=peer.difficulty,
                )

        case ObjectRequestType.OBJECT_PUT:
            node.logger.debug("Handling OBJECT_PUT for %s from %s", object_request.atom_id.hex(), peer.address)

            nearest_peer = node.peer_route.closest_peer_for_hash(object_request.atom_id)
            is_self_closest = False
            if nearest_peer is None or nearest_peer.address is None:
                is_self_closest = True
            else:
                try:
                    self_distance = xor_distance(object_request.atom_id, node.relay_public_key_bytes)
                    peer_distance = xor_distance(object_request.atom_id, nearest_peer.public_key_bytes)
                except Exception as exc:
                    node.logger.warning(
                        "Failed distance comparison for OBJECT_PUT %s: %s",
                        object_request.atom_id.hex(),
                        exc,
                    )
                    is_self_closest = True
                else:
                    is_self_closest = self_distance <= peer_distance

            if is_self_closest:
                node.logger.debug("Storing provider info for %s locally", object_request.atom_id.hex())
                provider_id = provider_id_for_payload(node, object_request.data)
                node.storage_index[object_request.atom_id] = provider_id
                print(
                    "OBJECT_PUT indexed provider atom_id=%s from=%s"
                    % (object_request.atom_id.hex(), peer.address)
                )
            else:
                node.logger.debug(
                    "Forwarding OBJECT_PUT for %s to nearer peer %s",
                    object_request.atom_id.hex(),
                    nearest_peer.address,
                )
                fwd_req = ObjectRequest(
                    type=ObjectRequestType.OBJECT_PUT,
                    data=object_request.data,
                    atom_id=object_request.atom_id,
                    payload_type=object_request.payload_type,
                )
                obj_req_msg = Message(
                    topic=MessageTopic.OBJECT_REQUEST,
                    body=fwd_req.to_bytes(),
                    sender=node.relay_public_key,
                )
                obj_req_msg.encrypt(nearest_peer.shared_key_bytes)
                enqueue_outgoing(
                    node,
                    nearest_peer.address,
                    message=obj_req_msg,
                    difficulty=nearest_peer.difficulty,
                )

        case _:
            node.logger.warning("Unknown ObjectRequestType %s from %s", object_request.type, peer.address)



from __future__ import annotations

import socket
from typing import Iterable, Tuple
from pathlib import Path

from cryptography.hazmat.primitives import serialization

from ..models.atom import Atom
from ..providers import provider_id_for_payload


def _hot_storage_set(self, key: bytes, value: Atom) -> bool:
    """Store atom in hot storage without exceeding the configured limit."""
    node_logger = self.logger
    projected = self.hot_storage_size + value.size
    hot_limit = self.config["hot_storage_limit"]
    if projected > hot_limit:
        node_logger.warning(
            "Hot storage limit reached (%s > %s); skipping atom %s",
            projected,
            hot_limit,
            key.hex(),
        )
        return False

    self.hot_storage[key] = value
    self.hot_storage_size = projected
    node_logger.debug(
        "Stored atom %s in hot storage (bytes=%s, total=%s)",
        key.hex(),
        value.size,
        projected,
    )
    return True


def _cold_storage_set(self, key: bytes, atom: Atom) -> None:
    """Persist an atom into the cold storage directory if it already exists."""
    node_logger = self.logger
    atom_hex = key.hex()
    if not self.config["cold_storage_path"]:
        node_logger.debug("Cold storage disabled; skipping atom %s", atom_hex)
        return
    atom_bytes = atom.to_bytes()
    projected = self.cold_storage_size + len(atom_bytes)
    cold_limit = self.config["cold_storage_limit"]
    if cold_limit and projected > cold_limit:
        node_logger.warning(
            "Cold storage limit reached (%s > %s); skipping atom %s",
            projected,
            cold_limit,
            atom_hex,
        )
        return
    directory = Path(self.config["cold_storage_path"])
    if not directory.exists():
        node_logger.warning(
            "Cold storage path %s missing; skipping atom %s",
            directory,
            atom_hex,
        )
        return
    filename = f"{atom_hex.upper()}.bin"
    file_path = directory / filename
    try:
        file_path.write_bytes(atom_bytes)
        self.cold_storage_size = projected
        node_logger.debug("Persisted atom %s to cold storage", atom_hex)
    except OSError as exc:
        node_logger.error(
            "Failed writing atom %s to cold storage %s: %s",
            atom_hex,
            file_path,
            exc,
        )


def _network_set(self, atom_id: bytes, payload_type: int) -> None:
    """Advertise an atom id to the closest known peer so they can fetch it from us."""
    node_logger = self.logger
    atom_hex = atom_id.hex()
    try:
        from ...communication.handlers.object_request import (
            ObjectRequest,
            ObjectRequestType,
        )
        from ...communication.models.message import Message, MessageTopic
        from ...communication.outgoing_queue import enqueue_outgoing
    except Exception as exc:
        node_logger.warning(
            "Communication module unavailable; cannot advertise atom %s: %s",
            atom_hex,
            exc,
        )
        return
    try:
        provider_ip, provider_port = self.incoming_socket.getsockname()[:2]
    except Exception as exc:
        node_logger.warning(
            "Unable to determine provider address for atom %s: %s",
            atom_hex,
            exc,
        )
        return

    try:
        provider_ip_bytes = socket.inet_aton(provider_ip)
        provider_port_bytes = int(provider_port).to_bytes(2, "big", signed=False)
        provider_key_bytes = self.relay_public_key_bytes
    except Exception as exc:
        node_logger.warning("Unable to encode provider info for %s: %s", atom_hex, exc)
        return

    provider_payload = provider_key_bytes + provider_ip_bytes + provider_port_bytes

    try:
        closest_peer = self.peer_route.closest_peer_for_hash(atom_id)
    except Exception as exc:
        node_logger.warning("Peer lookup failed for atom %s: %s", atom_hex, exc)
        return

    is_self_closest = False
    if closest_peer is None or closest_peer.address is None:
        is_self_closest = True
    else:
        try:
            from ...communication.util import xor_distance
        except Exception as exc:
            node_logger.warning("Failed to import xor_distance for atom %s: %s", atom_hex, exc)
            is_self_closest = True
        else:
            try:
                self_distance = xor_distance(atom_id, self.relay_public_key_bytes)
                peer_distance = xor_distance(atom_id, closest_peer.public_key_bytes)
            except Exception as exc:
                node_logger.warning("Failed computing distance for atom %s: %s", atom_hex, exc)
                is_self_closest = True
            else:
                is_self_closest = self_distance <= peer_distance

    if is_self_closest:
        node_logger.debug("Self is closest; indexing provider for atom %s", atom_hex)
        provider_id = provider_id_for_payload(self, provider_payload)
        self.storage_index[atom_id] = provider_id
        return

    target_addr = closest_peer.address

    obj_req = ObjectRequest(
        type=ObjectRequestType.OBJECT_PUT,
        data=provider_payload,
        atom_id=atom_id,
        payload_type=payload_type,
    )
    
    message_body = obj_req.to_bytes()

    message = Message(
        topic=MessageTopic.OBJECT_REQUEST,
        content=message_body,
        sender=self.relay_public_key,
    )
    message.encrypt(closest_peer.shared_key_bytes)
    try:
        queued = enqueue_outgoing(
            self,
            target_addr,
            message=message,
            difficulty=closest_peer.difficulty,
        )
        if queued:
            node_logger.debug(
                "Advertised atom %s to peer at %s:%s",
                atom_hex,
                target_addr[0],
                target_addr[1],
            )
        else:
            node_logger.debug(
                "Dropped atom advertisement %s to peer at %s:%s",
                atom_hex,
                target_addr[0],
                target_addr[1],
            )
    except Exception as exc:
        node_logger.error(
            "Failed to queue advertisement for atom %s to %s:%s: %s",
            atom_hex,
            target_addr[0],
            target_addr[1],
            exc,
        )


def add_atom_advertisement(
    self,
    atom_id: bytes,
    payload_type: int,
    expires_at: float | None = None,
) -> None:
    """Track an atom id for periodic advertisement."""
    entry = (atom_id, payload_type, expires_at)
    lock = getattr(self, "atom_advertisments_lock", None)
    if lock is None:
        self.atom_advertisments.append(entry)
        return
    with lock:
        self.atom_advertisments.append(entry)


def add_atom_advertisements(
    self,
    entries: Iterable[Tuple[bytes, int, float | None]],
) -> None:
    """Track multiple atom ids for periodic advertisement."""
    lock = getattr(self, "atom_advertisments_lock", None)
    if lock is None:
        self.atom_advertisments.extend(entries)
        return
    with lock:
        self.atom_advertisments.extend(entries)




from __future__ import annotations

from typing import Optional, Set, Any
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
from .block import Block
from ...storage.models.atom import ZERO32


class Fork:
    """A branch head within a Chain (same root).

    - head:       current tip block id (bytes)
    - peers:      identifiers (e.g., peer pubkey objects) following this head
    - root:       genesis block id for this chain (optional)
    - verified_up_to: earliest verified ancestor (optional)
    - chain_fork_position: the chain's fork anchor relevant to this fork
    """

    def __init__(
        self,
        *,
        head: bytes,
        verified_up_to: Optional[bytes] = None,
        chain_fork_position: Optional[bytes] = None,
        reached_genesis: bool = False,
        malicious_block_hash: Optional[bytes] = None,
    ) -> None:
        self.head: bytes = head
        self.peers: Set[Any] = set()
        self.root: Optional[bytes] = None
        self.verified_up_to: Optional[bytes] = verified_up_to
        self.chain_fork_position: Optional[bytes] = chain_fork_position
        self.reached_genesis: bool = bool(reached_genesis)
        # Mark the first block found malicious during validation; None means not found
        self.malicious_block_hash: Optional[bytes] = malicious_block_hash

    def to_dict(self) -> dict[str, Any]:
        return {
            "head": self.head,
            "verified_up_to": self.verified_up_to,
            "chain_fork_position": self.chain_fork_position,
            "reached_genesis": self.reached_genesis,
            "malicious_block_hash": self.malicious_block_hash,
        }

    @classmethod
    def from_dict(cls, payload: dict[str, Any]) -> Fork:
        head = payload.get("head")
        if not isinstance(head, (bytes, bytearray)):
            raise TypeError("fork head must be bytes")

        verified_up_to = payload.get("verified_up_to")
        if verified_up_to is not None and not isinstance(
            verified_up_to, (bytes, bytearray)
        ):
            raise TypeError("fork verified_up_to must be bytes or None")

        chain_fork_position = payload.get("chain_fork_position")
        if chain_fork_position is not None and not isinstance(
            chain_fork_position, (bytes, bytearray)
        ):
            raise TypeError("fork chain_fork_position must be bytes or None")

        reached_genesis = payload.get("reached_genesis")
        if not isinstance(reached_genesis, bool):
            raise TypeError("fork reached_genesis must be bool")

        malicious_block_hash = payload.get("malicious_block_hash")
        if malicious_block_hash is not None and not isinstance(
            malicious_block_hash, (bytes, bytearray)
        ):
            raise TypeError("fork malicious_block_hash must be bytes or None")

        return cls(
            head=bytes(head),
            verified_up_to=bytes(verified_up_to)
            if isinstance(verified_up_to, (bytes, bytearray))
            else None,
            chain_fork_position=bytes(chain_fork_position)
            if isinstance(chain_fork_position, (bytes, bytearray))
            else None,
            reached_genesis=reached_genesis,
            malicious_block_hash=bytes(malicious_block_hash)
            if isinstance(malicious_block_hash, (bytes, bytearray))
            else None,
        )

    def add_peer(self, peer_id: Any) -> None:
        self.peers.add(peer_id)

    def remove_peer(self, peer_id: Any) -> None:
        self.peers.discard(peer_id)

    def verify(self, node: Any) -> bool:
        """Verify this fork using the node to manage fork splits/joins."""
        def _hex(value: Optional[bytes]) -> str:
            if isinstance(value, (bytes, bytearray)):
                return value.hex()
            return str(value)

        node.logger.debug("Fork verify start head=%s", _hex(self.head))

        visited_set: Set[bytes] = set()
        anchor_hash: Optional[bytes] = None
        anchor_kind: Optional[str] = None
        intersection_fork_head: Optional[bytes] = None
        anchor_validated = False

        def validate_header(child: Block, parent: Optional[Block]) -> bool:
            """
            Lightweight/header validation without tx/receipt/account checks.

            The caller supplies the parent block (or None for genesis) so we can
            verify linkage, height, timestamps, and difficulty in a single pass.
            """
            is_genesis = parent is None or (child.previous_block_hash or ZERO32) == ZERO32

            node_chain = getattr(node, "chain", None)
            if node_chain is not None and child.chain_id != node_chain:
                node.logger.debug(
                    "Header verify failed chain_id=%s expected=%s block=%s",
                    child.chain_id,
                    node_chain,
                    _hex(child.atom_hash),
                )
                return False

            # Basic field presence
            if child.timestamp is None:
                node.logger.debug(
                    "Header verify failed missing timestamp block=%s",
                    _hex(child.atom_hash),
                )
                return False
            if not is_genesis:
                if not child.body_hash or not child.signature or not child.validator_public_key_bytes:
                    node.logger.debug(
                        "Header verify failed missing body/signature/validator block=%s",
                        _hex(child.atom_hash),
                    )
                    return False

            # Linkage rules
            if is_genesis:
                if (child.previous_block_hash or ZERO32) != ZERO32:
                    node.logger.debug(
                        "Header verify failed genesis prev_hash=%s block=%s",
                        _hex(child.previous_block_hash),
                        _hex(child.atom_hash),
                    )
                    return False
                if child.number not in (0,):
                    node.logger.debug(
                        "Header verify failed genesis number=%s block=%s",
                        child.number,
                        _hex(child.atom_hash),
                    )
                    return False
            else:
                parent_hash = parent.atom_hash or ZERO32
                if (child.previous_block_hash or ZERO32) != parent_hash:
                    node.logger.debug(
                        "Header verify failed prev hash mismatch block=%s prev=%s expected=%s",
                        _hex(child.atom_hash),
                        _hex(child.previous_block_hash),
                        _hex(parent_hash),
                    )
                    return False
                expected_number = (parent.number or 0) + 1
                if child.number != expected_number:
                    node.logger.debug(
                        "Header verify failed number mismatch block=%s number=%s expected=%s",
                        _hex(child.atom_hash),
                        child.number,
                        expected_number,
                    )
                    return False

                parent_ts = parent.timestamp
                if parent_ts is not None and int(child.timestamp) < int(parent_ts) + 1:
                    node.logger.debug(
                        "Header verify failed timestamp block=%s ts=%s parent_ts=%s",
                        _hex(child.atom_hash),
                        child.timestamp,
                        parent_ts,
                    )
                    return False

                # Signature over body hash
                try:
                    pub = Ed25519PublicKey.from_public_bytes(
                        bytes(child.validator_public_key_bytes)
                    )
                    pub.verify(child.signature, child.body_hash)  # type: ignore[arg-type]
                except InvalidSignature:
                    node.logger.debug(
                        "Header verify failed signature block=%s",
                        _hex(child.atom_hash),
                    )
                    return False
                except Exception:
                    node.logger.debug(
                        "Header verify failed signature error block=%s",
                        _hex(child.atom_hash),
                    )
                    return False

                # Difficulty and PoW
                expected_diff = Block.calculate_delay_difficulty(
                    previous_timestamp=parent.timestamp,
                    current_timestamp=child.timestamp,
                    previous_difficulty=parent.delay_difficulty,
                )
                if child.delay_difficulty is None or int(child.delay_difficulty) != int(
                    expected_diff
                ):
                    node.logger.debug(
                        "Header verify failed difficulty block=%s diff=%s expected=%s",
                        _hex(child.atom_hash),
                        child.delay_difficulty,
                        expected_diff,
                    )
                    return False

                required_work = max(1, int(parent.delay_difficulty or 1))
                block_hash = child.atom_hash or b""
                if not block_hash:
                    node.logger.debug(
                        "Header verify failed missing hash block=%s",
                        _hex(child.atom_hash),
                    )
                    return False
                if Block._leading_zero_bits(block_hash) < required_work:
                    node.logger.debug(
                        "Header verify failed pow block=%s zeros=%s required=%s",
                        _hex(child.atom_hash),
                        Block._leading_zero_bits(block_hash),
                        required_work,
                    )
                    return False

            return True

        def is_on_other_fork_path(target_hash: bytes) -> Optional[bytes]:
            """Return the head of a fork whose ancestry includes target_hash."""
            for other_head in node.forks:
                if other_head == self.head:
                    continue
                blk_hash = other_head
                seen: Set[bytes] = set()
                while blk_hash and blk_hash not in seen:
                    seen.add(blk_hash)
                    if blk_hash == target_hash:
                        return other_head
                    try:
                        blk = Block.from_atom(node, blk_hash)
                    except Exception:
                        node.logger.debug(
                            "Fork path lookup failed loading block=%s",
                            _hex(blk_hash),
                        )
                        blk = None
                    if blk is None:
                        break
                    prev = getattr(blk, "previous_block_hash", ZERO32) or ZERO32
                    if prev == ZERO32:
                        break
                    blk_hash = prev
            return None

        cursor = self.head
        pending_child: Optional[Block] = None
        while cursor and cursor not in visited_set:
            try:
                blk = Block.from_atom(node, cursor)
            except Exception:
                node.logger.debug("Fork verify failed loading block=%s", _hex(cursor))
                blk = None
            if blk is None:
                self.malicious_block_hash = (
                    pending_child.atom_hash if pending_child else cursor
                )
                node.logger.warning(
                    "Fork verify failed missing block=%s pending=%s",
                    _hex(cursor),
                    _hex(pending_child.atom_hash) if pending_child else None,
                )
                return False

            if pending_child is not None:
                if not validate_header(pending_child, blk):
                    self.malicious_block_hash = (
                        pending_child.atom_hash
                        or pending_child.body_hash
                        or pending_child.previous_block_hash
                        or cursor
                    )
                    node.logger.warning(
                        "Fork verify failed header block=%s parent=%s",
                        _hex(pending_child.atom_hash),
                        _hex(blk.atom_hash),
                    )
                    return False
                if not pending_child.atom_hash:
                    self.malicious_block_hash = (
                        pending_child.body_hash
                        or pending_child.previous_block_hash
                        or cursor
                    )
                    node.logger.warning(
                        "Fork verify failed missing hash block=%s",
                        _hex(pending_child.body_hash),
                    )
                    return False
                if anchor_hash is not None and pending_child.atom_hash == anchor_hash:
                    anchor_validated = True
                    node.logger.debug(
                        "Fork verify reached anchor=%s kind=%s",
                        _hex(anchor_hash),
                        anchor_kind,
                    )
                    break

            visited_set.add(cursor)

            if anchor_hash is None:
                if cursor in node.forks and cursor != self.head:
                    anchor_hash = cursor
                    anchor_kind = "fork_head"
                    node.logger.debug(
                        "Fork verify anchor fork_head=%s",
                        _hex(anchor_hash),
                    )
                else:
                    other_head = is_on_other_fork_path(cursor)
                    if other_head:
                        anchor_hash = cursor
                        anchor_kind = "intersection"
                        intersection_fork_head = other_head
                        node.logger.debug(
                            "Fork verify anchor intersection=%s other_head=%s",
                            _hex(anchor_hash),
                            _hex(other_head),
                        )
                    else:
                        prev_hash = getattr(blk, "previous_block_hash", ZERO32) or ZERO32
                        if prev_hash == ZERO32:
                            self.reached_genesis = True
                            anchor_hash = cursor
                            anchor_kind = "genesis"
                            node.logger.debug(
                                "Fork verify anchor genesis=%s",
                                _hex(anchor_hash),
                            )

            pending_child = blk
            prev_hash = getattr(blk, "previous_block_hash", ZERO32) or ZERO32
            if prev_hash == ZERO32:
                self.reached_genesis = True
                break
            cursor = prev_hash

        if pending_child is not None and not anchor_validated:
            parent_blk: Optional[Block] = None
            prev_hash = getattr(pending_child, "previous_block_hash", ZERO32) or ZERO32
            if prev_hash not in (None, ZERO32, b""):
                try:
                    parent_blk = Block.from_atom(node, prev_hash)
                except Exception:
                    node.logger.debug(
                        "Fork verify failed loading parent block=%s",
                        _hex(prev_hash),
                    )
                    parent_blk = None
            if not validate_header(pending_child, parent_blk):
                self.malicious_block_hash = (
                    pending_child.atom_hash
                    or pending_child.body_hash
                    or pending_child.previous_block_hash
                    or self.head
                )
                node.logger.warning(
                    "Fork verify failed header block=%s parent=%s",
                    _hex(pending_child.atom_hash),
                    _hex(parent_blk.atom_hash) if parent_blk else None,
                )
                return False
            if not pending_child.atom_hash:
                self.malicious_block_hash = (
                    pending_child.body_hash
                    or pending_child.previous_block_hash
                    or self.head
                )
                node.logger.warning(
                    "Fork verify failed missing hash block=%s",
                    _hex(pending_child.body_hash),
                )
                return False
            if anchor_hash is None:
                anchor_hash = pending_child.atom_hash
                anchor_kind = "genesis"
                self.reached_genesis = True
                node.logger.debug(
                    "Fork verify anchor genesis=%s",
                    _hex(anchor_hash),
                )
            if pending_child.atom_hash == anchor_hash:
                anchor_validated = True

        if anchor_hash is None or not anchor_validated:
            node.logger.warning(
                "Fork verify failed anchor validated=%s anchor=%s",
                anchor_validated,
                _hex(anchor_hash),
            )
            return False

        node.logger.debug(
            "Fork verify heavy pass head=%s anchor=%s",
            _hex(self.head),
            _hex(anchor_hash),
        )
        heavy_cursor = self.head
        heavy_pending: Optional[Block] = None
        heavy_seen: Set[bytes] = set()
        heavy_anchor_verified = False
        while heavy_cursor and heavy_cursor not in heavy_seen:
            heavy_seen.add(heavy_cursor)
            try:
                blk = Block.from_atom(node, heavy_cursor)
            except Exception:
                self.malicious_block_hash = (
                    heavy_pending.atom_hash if heavy_pending else heavy_cursor
                )
                node.logger.warning(
                    "Fork verify failed heavy load block=%s pending=%s",
                    _hex(heavy_cursor),
                    _hex(heavy_pending.atom_hash) if heavy_pending else None,
                )
                return False

            if heavy_pending is not None:
                heavy_pending.previous_block = blk
                if not heavy_pending.verify(node):
                    self.malicious_block_hash = (
                        heavy_pending.atom_hash
                        or heavy_pending.previous_block_hash
                        or heavy_cursor
                    )
                    node.logger.warning(
                        "Fork verify failed heavy block=%s parent=%s",
                        _hex(heavy_pending.atom_hash),
                        _hex(blk.atom_hash),
                    )
                    return False
                if heavy_pending.atom_hash == anchor_hash:
                    heavy_anchor_verified = True
                    node.logger.debug(
                        "Fork verify heavy reached anchor=%s",
                        _hex(anchor_hash),
                    )
                    break

            prev_hash = getattr(blk, "previous_block_hash", ZERO32) or ZERO32
            heavy_pending = blk
            if prev_hash == ZERO32:
                break
            heavy_cursor = prev_hash

        if not heavy_anchor_verified and heavy_pending is not None:
            if heavy_pending.atom_hash == anchor_hash:
                heavy_pending.previous_block = None
                if not heavy_pending.verify(node):
                    self.malicious_block_hash = (
                        heavy_pending.atom_hash
                        or heavy_pending.previous_block_hash
                        or self.head
                    )
                    node.logger.warning(
                        "Fork verify failed heavy anchor block=%s",
                        _hex(heavy_pending.atom_hash),
                    )
                    return False
                heavy_anchor_verified = True

        if not heavy_anchor_verified:
            node.logger.warning(
                "Fork verify failed heavy anchor verified=%s anchor=%s",
                heavy_anchor_verified,
                _hex(anchor_hash),
            )
            return False

        # Commit staged fork edits
        if anchor_kind == "fork_head":
            ref = node.forks.get(anchor_hash)
            chain_anchor = ref.chain_fork_position if ref else anchor_hash
            base_root = ref.root if ref and ref.root else anchor_hash
            self.verified_up_to = anchor_hash
            self.chain_fork_position = chain_anchor or anchor_hash
            self.root = base_root
            self.malicious_block_hash = None
            node.forks[self.head] = self
            node.logger.debug(
                "Fork verify committed fork_head head=%s anchor=%s",
                _hex(self.head),
                _hex(anchor_hash),
            )
            return True

        if anchor_kind == "intersection":
            base_root = anchor_hash
            existing = node.forks.get(intersection_fork_head) if intersection_fork_head else None
            if existing and existing.root:
                base_root = existing.root

            base_fork = node.forks.get(anchor_hash)
            if base_fork is None:
                base_fork = Fork(head=anchor_hash)
            base_fork.root = base_root
            base_fork.chain_fork_position = anchor_hash
            base_fork.verified_up_to = anchor_hash

            if existing is not None:
                existing.chain_fork_position = anchor_hash
                existing.verified_up_to = anchor_hash
                existing.root = base_root
                node.forks[existing.head] = existing

            self.chain_fork_position = anchor_hash
            self.verified_up_to = anchor_hash
            self.root = base_root
            self.malicious_block_hash = None

            node.forks[base_fork.head] = base_fork
            node.forks[self.head] = self
            node.logger.debug(
                "Fork verify committed intersection head=%s anchor=%s",
                _hex(self.head),
                _hex(anchor_hash),
            )
            return True

        if anchor_kind == "genesis":
            self.verified_up_to = anchor_hash
            self.chain_fork_position = anchor_hash
            self.root = anchor_hash
            self.malicious_block_hash = None
            node.forks[self.head] = self
            node.logger.debug(
                "Fork verify committed genesis head=%s anchor=%s",
                _hex(self.head),
                _hex(anchor_hash),
            )
            return True

        return False



from __future__ import annotations

import time
from queue import Empty
from typing import Any, Optional, Set, Tuple

from astreum.validation.models.fork import Fork
from astreum.validation.models.block import Block


def _process_peers_latest_block(
    node: Any, latest_block_hash: bytes, peer_ids: Set[Any]
) -> None:
    """Assign peers to the fork that matches their reported head."""
    node.logger.debug(
        "Processing %d peers reporting block %s",
        len(peer_ids),
        latest_block_hash.hex()
        if isinstance(latest_block_hash, (bytes, bytearray))
        else latest_block_hash,
    )
    new_fork = Fork(head=latest_block_hash)

    new_fork.verify(node)

    if new_fork.verified_up_to and new_fork.verified_up_to in node.forks:
        ref = node.forks[new_fork.verified_up_to]
        if getattr(ref, "malicious_block_hash", None):
            node.logger.warning(
                "Skipping fork from block %s referencing malicious fork %s",
                latest_block_hash.hex()
                if isinstance(latest_block_hash, (bytes, bytearray))
                else latest_block_hash,
            new_fork.verified_up_to.hex()
            if isinstance(new_fork.verified_up_to, (bytes, bytearray))
            else new_fork.verified_up_to,
            )
            return
        new_fork.root = ref.root
        new_fork.verified_up_to = ref.verified_up_to
        new_fork.chain_fork_position = ref.chain_fork_position

    for peer_id in peer_ids:
        new_fork.add_peer(peer_id)
        for head, fork in list(node.forks.items()):
            if head != latest_block_hash:
                fork.remove_peer(peer_id)

    node.forks[latest_block_hash] = new_fork
    node.logger.debug(
        "Fork %s now has %d peers (total forks %d)",
        latest_block_hash.hex()
        if isinstance(latest_block_hash, (bytes, bytearray))
        else latest_block_hash,
        len(new_fork.peers),
        len(node.forks),
    )


def _select_best_fork_head(node: Any) -> Optional[Tuple[bytes, Block, int]]:
    forks = getattr(node, "forks", None)
    if not isinstance(forks, dict) or not forks:
        return None

    config = getattr(node, "config", {}) or {}
    try:
        max_stale = int(config.get("verification_max_stale_seconds", 10))
    except (TypeError, ValueError):
        max_stale = 10
    try:
        max_future = int(config.get("verification_max_future_skew", 2))
    except (TypeError, ValueError):
        max_future = 2

    now = int(time.time())
    current_head = getattr(node, "latest_block_hash", None)

    best_head: Optional[bytes] = None
    best_block: Optional[Block] = None
    best_height: int = -1

    for head, fork in list(forks.items()):
        if getattr(fork, "malicious_block_hash", None):
            continue
        if not getattr(fork, "verified_up_to", None):
            continue
        if not getattr(fork, "peers", None):
            continue
        if not isinstance(head, (bytes, bytearray)):
            continue

        try:
            block = Block.from_atom(node, head)
        except Exception:
            continue

        ts = getattr(block, "timestamp", None)
        if ts is None:
            continue
        ts_int = int(ts)
        if max_stale >= 0 and (now - ts_int) > max_stale:
            continue
        if max_future >= 0 and (ts_int - now) > max_future:
            continue

        height = int(getattr(block, "number", 0) or 0)
        if height > best_height:
            best_head = bytes(head)
            best_block = block
            best_height = height
            continue
        if height == best_height:
            if current_head == head:
                best_head = bytes(head)
                best_block = block
                best_height = height
            elif current_head != best_head and best_head is not None:
                if bytes(head) < bytes(best_head):
                    best_head = bytes(head)
                    best_block = block
                    best_height = height

    if best_head is None or best_block is None:
        return None
    return best_head, best_block, best_height


def make_verify_worker(node: Any):
    """Build the verify worker bound to the given node."""

    def _verify_worker() -> None:
        node.logger.info("Verify worker started")
        stop = node._verify_stop_event
        while not stop.is_set():
            batch: list[tuple[bytes, Set[Any]]] = []
            try:
                while True:
                    latest_b, peers = node._validation_verify_queue.get_nowait()
                    batch.append((latest_b, peers))
            except Empty:
                pass

            if not batch:
                node.logger.debug("Verify queue empty; sleeping")
                time.sleep(0.1)
                continue

            batch.sort(key=lambda item: len(item[1]), reverse=True)

            for latest_b, peers in batch:
                try:
                    _process_peers_latest_block(node, latest_b, peers)
                    node.logger.debug(
                        "Updated forks from block %s for %d peers",
                        latest_b.hex()
                        if isinstance(latest_b, (bytes, bytearray))
                        else latest_b,
                        len(peers),
                    )
                except Exception:
                    latest_hex = (
                        latest_b.hex()
                        if isinstance(latest_b, (bytes, bytearray))
                        else latest_b
                    )
                    node.logger.exception(
                        "Failed processing verification batch for %s", latest_hex
                    )

            selected = _select_best_fork_head(node)
            if selected is not None:
                selected_head, selected_block, selected_height = selected
                if getattr(node, "latest_block_hash", None) != selected_head:
                    node.latest_block_hash = selected_head
                    node.latest_block = selected_block
                    node.logger.info(
                        "Selected verified head %s (height=%s)",
                        selected_head.hex(),
                        selected_height,
                    )
        node.logger.info("Verify worker stopped")

    return _verify_worker



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

from ..models.ping import Ping
from ..models.peer import Peer

if TYPE_CHECKING:
    from .... import Node


def handle_ping(node: "Node", peer: Peer, payload: bytes) -> None:
    """Update peer and validation state based on an incoming ping message."""
    try:
        ping = Ping.from_bytes(payload)
    except Exception as exc:
        node.logger.warning("Error decoding ping: %s", exc)
        return

    peer.timestamp = datetime.now(timezone.utc)
    peer.latest_block = ping.latest_block
    peer.difficulty = ping.difficulty
    if peer.is_default_seed and ping.latest_block:
        if getattr(node, "latest_block_hash", None) != ping.latest_block:
            node.latest_block_hash = ping.latest_block
            node.latest_block = None
            try:
                from astreum.validation.models.block import Block

                node.latest_block = Block.from_atom(node, ping.latest_block)
                node.logger.info(
                    "Loaded latest block %s after seed update",
                    ping.latest_block.hex(),
                )
            except Exception as exc:
                node.logger.warning(
                    "Failed loading latest block %s after seed update: %s",
                    ping.latest_block.hex(),
                    exc,
                )
            node.logger.info(
                "Updated latest block hash from default seed %s",
                peer.address[0] if peer.address else "unknown",
            )

    validation_route = node.validation_route
    if validation_route is None:
        return

    try:
        if ping.is_validator:
            validation_route.add_peer(peer.public_key_bytes)
        else:
            validation_route.remove_peer(peer.public_key_bytes)
    except Exception:
        pass



"""Core Astreum Node implementation."""

from __future__ import annotations

import threading
import uuid
from typing import Dict

from astreum.communication.node import connect_node
from astreum.communication.util import get_bootstrap_peers
from astreum.communication.disconnect import disconnect_node
from astreum.communication.models.peer import (
    add_peer as peers_add_peer,
    replace_peer as peers_replace_peer,
    get_peer as peers_get_peer,
    remove_peer as peers_remove_peer,
)
from astreum.validation.node import validate_blockchain
from astreum.verification.node import verify_blockchain
from astreum.machine import Expr, high_eval, low_eval, script_eval
from astreum.machine.models.environment import Env, env_get, env_set
from astreum.machine.models.expression import get_expr_list_from_storage
from astreum.storage.actions.get import (
    _hot_storage_get,
    _cold_storage_get,
    _network_get,
    get_atom_from_local_storage,
    get_atom,
    get_atom_list_from_local_storage,
    get_atom_list,
)
from astreum.storage.actions.set import (
    _hot_storage_set,
    _cold_storage_set,
    _network_set,
    add_atom_advertisement,
    add_atom_advertisements,
)
from astreum.storage.requests import add_atom_req, has_atom_req, pop_atom_req
from astreum.storage.setup import storage_setup
from astreum.utils.config import config_setup
from astreum.utils.logging import logging_setup


class Node:
    def __init__(self, config: dict = {}):
        self.config = config_setup(config=config)
        self.bootstrap_peers = get_bootstrap_peers(self)
        
        self.logger = logging_setup(self.config)

        self.logger.info("Starting Astreum Node")

        # Chain Configuration
        self.logger.info(f"Chain configured as: {self.config["chain"]} ({self.config["chain_id"]})")

        # Storage Setup
        storage_setup(self, config=self.config)

        # Machine Setup
        self.environments: Dict[uuid.UUID, Env] = {}
        self.machine_environments_lock = threading.RLock()
        self.is_connected = False
        self.latest_block_hash = None
        self.latest_block = None
        
    connect = connect_node
    disconnect = disconnect_node

    verify = verify_blockchain

    validate = validate_blockchain

    low_eval = low_eval
    high_eval = high_eval
    script_eval = script_eval

    env_get = env_get
    env_set = env_set

    # Storage
    ## Get
    _hot_storage_get = _hot_storage_get
    _cold_storage_get = _cold_storage_get
    _network_get = _network_get

    ## Set
    _hot_storage_set = _hot_storage_set
    _cold_storage_set = _cold_storage_set
    _network_set = _network_set
    add_atom_advertisement = add_atom_advertisement
    add_atom_advertisements = add_atom_advertisements

    get_atom_from_local_storage = get_atom_from_local_storage
    get_atom = get_atom
    get_atom_list_from_local_storage = get_atom_list_from_local_storage
    get_atom_list = get_atom_list

    get_expr_list_from_storage = get_expr_list_from_storage

    add_atom_req = add_atom_req
    has_atom_req = has_atom_req
    pop_atom_req = pop_atom_req

    add_peer = peers_add_peer
    replace_peer = peers_replace_peer
    get_peer = peers_get_peer
    remove_peer = peers_remove_peer



from __future__ import annotations

from typing import Optional



def provider_id_for_payload(node, payload: bytes) -> int:
    """Return the provider id for a payload, inserting if new."""
    for idx, existing in enumerate(node.storage_providers):
        if existing == payload:
            return idx
    node.storage_providers.append(payload)
    return len(node.storage_providers) - 1



def provider_payload_for_id(node, provider_id: int) -> Optional[bytes]:
    """Return the provider payload for a provider id, or None."""
    if not isinstance(provider_id, int) or provider_id < 0:
        return None
    try:
        return node.storage_providers[provider_id]
    except IndexError:
        return None



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, List, Optional, Tuple

from ...storage.models.atom import Atom, AtomKind, ZERO32
from ...utils.integer import bytes_to_int, int_to_bytes
from .account import Account
from ..constants import TREASURY_ADDRESS
from .receipt import STATUS_FAILED, Receipt, STATUS_SUCCESS

@dataclass
class Transaction:
    chain_id: int
    amount: int
    counter: int
    version: int = 1
    data: bytes = b""
    recipient: bytes = b""
    sender: bytes = b""
    signature: bytes = b""
    hash: bytes = ZERO32

    def sign(self, private_key: Any) -> bytes:
        """Sign the transaction detail list head and store the signature."""
        detail_payloads: List[bytes] = []

        def emit(payload: bytes) -> None:
            detail_payloads.append(payload)

        emit(int_to_bytes(self.chain_id))
        emit(int_to_bytes(self.amount))
        emit(int_to_bytes(self.counter))
        emit(bytes(self.data))
        emit(bytes(self.recipient))
        emit(bytes(self.sender))

        body_head = ZERO32
        for payload in reversed(detail_payloads):
            atom = Atom(data=payload, next_id=body_head, kind=AtomKind.BYTES)
            body_head = atom.object_id()

        self.signature = private_key.sign(body_head)
        return body_head

    def to_atom(self) -> Tuple[bytes, List[Atom]]:
        """Serialise the transaction, returning (object_id, atoms)."""
        detail_payloads: List[bytes] = []
        acc: List[Atom] = []

        def emit(payload: bytes) -> None:
            detail_payloads.append(payload)

        emit(int_to_bytes(self.chain_id))
        emit(int_to_bytes(self.amount))
        emit(int_to_bytes(self.counter))
        emit(bytes(self.data))
        emit(bytes(self.recipient))
        emit(bytes(self.sender))

        body_head = ZERO32
        detail_atoms: List[Atom] = []
        for payload in reversed(detail_payloads):
            atom = Atom(data=payload, next_id=body_head, kind=AtomKind.BYTES)
            detail_atoms.append(atom)
            body_head = atom.object_id()
        detail_atoms.reverse()
        acc.extend(detail_atoms)

        body_list_atom = Atom(data=body_head, kind=AtomKind.LIST)
        acc.append(body_list_atom)
        body_list_id = body_list_atom.object_id()

        signature_atom = Atom(
            data=bytes(self.signature),
            next_id=body_list_id,
            kind=AtomKind.BYTES,
        )
        version_atom = Atom(
            data=int_to_bytes(self.version),
            next_id=signature_atom.object_id(),
            kind=AtomKind.BYTES,
        )
        type_atom = Atom(
            data=b"transaction",
            next_id=version_atom.object_id(),
            kind=AtomKind.SYMBOL,
        )

        acc.append(signature_atom)
        acc.append(version_atom)
        acc.append(type_atom)

        self.hash = type_atom.object_id()
        return self.hash, acc

    @classmethod
    def from_atom(
        cls,
        node: Any,
        transaction_id: bytes,
    ) -> Transaction:
        get_atom = getattr(node, "get_atom", None)
        if not callable(get_atom):
            raise NotImplementedError("node does not expose an atom getter")

        def _atom_kind(atom: Optional[Atom]) -> Optional[AtomKind]:
            kind_value = getattr(atom, "kind", None)
            if isinstance(kind_value, AtomKind):
                return kind_value
            if isinstance(kind_value, int):
                try:
                    return AtomKind(kind_value)
                except ValueError:
                    return None
            return None

        def _require_atom(
            atom_id: Optional[bytes],
            context: str,
            expected_kind: Optional[AtomKind] = None,
        ) -> Atom:
            if not atom_id or atom_id == ZERO32:
                raise ValueError(f"missing {context}")
            atom = get_atom(atom_id)
            if atom is None:
                raise ValueError(f"missing {context}")
            if expected_kind is not None:
                kind = _atom_kind(atom)
                if kind is not expected_kind:
                    raise ValueError(f"malformed {context}")
            return atom

        type_atom = _require_atom(transaction_id, "transaction type atom", AtomKind.SYMBOL)
        if type_atom.data != b"transaction":
            raise ValueError("not a transaction (type atom payload)")

        version_atom = _require_atom(type_atom.next_id, "transaction version atom", AtomKind.BYTES)
        version = bytes_to_int(version_atom.data)
        if version != 1:
            raise ValueError("unsupported transaction version")

        signature_atom = _require_atom(
            version_atom.next_id,
            "transaction signature atom",
            AtomKind.BYTES,
        )
        body_list_atom = _require_atom(signature_atom.next_id, "transaction body list atom", AtomKind.LIST)
        if body_list_atom.next_id and body_list_atom.next_id != ZERO32:
            raise ValueError("malformed transaction (body list tail)")

        detail_atoms = node.get_atom_list(body_list_atom.data)
        if detail_atoms is None:
            raise ValueError("missing transaction body list nodes")
        if len(detail_atoms) != 6:
            raise ValueError("transaction body must contain exactly 6 detail entries")

        detail_values: List[bytes] = []
        for detail_atom in detail_atoms:
            if detail_atom.kind is not AtomKind.BYTES:
                raise ValueError("transaction detail atoms must be bytes")
            detail_values.append(detail_atom.data)

        (
            chain_id_bytes,
            amount_bytes,
            counter_bytes,
            data_bytes,
            recipient_bytes,
            sender_bytes,
        ) = detail_values

        return cls(
            chain_id=bytes_to_int(chain_id_bytes),
            amount=bytes_to_int(amount_bytes),
            counter=bytes_to_int(counter_bytes),
            data=data_bytes,
            recipient=recipient_bytes,
            sender=sender_bytes,
            signature=signature_atom.data,
            hash=bytes(transaction_id),
            version=version,
        )

    @classmethod
    def get_atoms(
        cls,
        node: Any,
        transaction_id: bytes,
    ) -> Optional[List[Atom]]:
        """Load the transaction atom chain from storage, returning the atoms or None."""
        atoms = node.get_atom_list(transaction_id)
        if atoms is None or len(atoms) < 4:
            return None
        type_atom = atoms[0]
        if type_atom.kind is not AtomKind.SYMBOL or type_atom.data != b"transaction":
            return None
        version_atom = atoms[1]
        if version_atom.kind is not AtomKind.BYTES or bytes_to_int(version_atom.data) != 1:
            return None

        body_list_atom = atoms[-1]
        detail_atoms = node.get_atom_list(body_list_atom.data)
        if detail_atoms is None:
            return None
        atoms.extend(detail_atoms)

        return atoms


def apply_transaction(node: Any, block: object, transaction_hash: bytes) -> int:
    """Apply transaction to the candidate block and return the collected fee."""
    transaction = Transaction.from_atom(node, transaction_hash)

    block_chain = getattr(block, "chain_id", None)
    if block_chain is not None and transaction.chain_id != block_chain:
        return 0

    accounts = getattr(block, "accounts", None)
    if accounts is None:
        raise ValueError("block missing accounts snapshot for transaction application")

    sender_account = accounts.get_account(address=transaction.sender, node=node)
    if sender_account is None:
        return 0

    tx_fee = 1
    tx_cost = tx_fee + transaction.amount

    if sender_account.balance < tx_cost:
        low_sender_balance_receipt = Receipt(
            transaction_hash=bytes(transaction_hash),
            cost=0,
            status=STATUS_FAILED,
        )
        low_sender_balance_receipt.to_atom()
        if block.receipts is None:
            block.receipts = []
        block.receipts.append(low_sender_balance_receipt)
        if block.transactions is None:
            block.transactions = []
        block.transactions.append(transaction)
        return 0

    recipient_account = accounts.get_account(address=transaction.recipient, node=node)
    if recipient_account is None:
        recipient_account = Account.create()

    if transaction.recipient == TREASURY_ADDRESS:
        stake_trie = recipient_account.data
        existing_stake = stake_trie.get(node, transaction.sender)
        current_stake = bytes_to_int(existing_stake)
        new_stake = current_stake + transaction.amount
        stake_trie.put(node, transaction.sender, int_to_bytes(new_stake))
        recipient_account.data_hash = stake_trie.root_hash or ZERO32
        recipient_account.balance += transaction.amount
    else:
        recipient_account.balance += transaction.amount

    sender_account.balance -= tx_cost
    accounts.set_account(transaction.sender, sender_account)
    accounts.set_account(transaction.recipient, recipient_account)

    if block.transactions is None:
        block.transactions = []
    block.transactions.append(transaction)

    receipt = Receipt(
        transaction_hash=bytes(transaction_hash),
        cost=tx_fee,
        status=STATUS_SUCCESS,
    )
    receipt.to_atom()
    if block.receipts is None:
        block.receipts = []
    block.receipts.append(receipt)
    return tx_fee

def send_transaction(
    node: "Node",
    receipient_public_key: bytes,
    sender_secret_key: bytes,
    amount: int
):
    if not getattr(node, "is_connected", False):
        raise RuntimeError("node not connected")

    latest_block = getattr(node, "latest_block", None)
    if latest_block is None:
        raise RuntimeError("latest block unavailable")

    import time
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
    from ...communication.handlers.object_response import OBJECT_FOUND_LIST_PAYLOAD
    from ...communication.models.message import Message, MessageTopic
    from ...communication.outgoing_queue import enqueue_outgoing
    from ...storage.advertisments import advertise_atoms
    from .accounts import Accounts

    sender_key = Ed25519PrivateKey.from_private_bytes(bytes(sender_secret_key))
    sender_public_key_bytes = sender_key.public_key().public_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PublicFormat.Raw,
    )

    accounts = Accounts(root_hash=getattr(latest_block, "accounts_hash", None))
    sender_account = accounts.get_account(address=sender_public_key_bytes, node=node)
    sender_counter = sender_account.counter if sender_account is not None else 0

    transaction = Transaction(
        chain_id=int(node.config.get("chain_id", 0)),
        amount=int(amount),
        counter=sender_counter + 1,
        data=b"",
        recipient=bytes(receipient_public_key),
        sender=sender_public_key_bytes,
    )
    body_head = transaction.sign(sender_key)
    tx_hash, atoms = transaction.to_atom()

    for atom in atoms:
        atom_id = atom.object_id()
        node._hot_storage_set(atom_id, atom)
        node._cold_storage_set(atom_id, atom)

    ttl_seconds = int(node.config["peer_timeout"])
    expires_at = time.time() + ttl_seconds if ttl_seconds > 0 else None
    entries = []
    for atom_id in (tx_hash, body_head):
        if atom_id and atom_id != ZERO32:
            entries.append((atom_id, OBJECT_FOUND_LIST_PAYLOAD, expires_at))
    if entries:
        node.add_atom_advertisements(entries)
        advertise_atoms(node, entries=entries)

    validation_route = getattr(node, "validation_route", None)
    if validation_route is None:
        raise RuntimeError("no validator available")

    has_validators = bool(getattr(validation_route, "peers", None))
    if not has_validators:
        peers = []
        peers_lock = getattr(node, "peers_lock", None)
        if peers_lock is not None:
            with peers_lock:
                peers = list(getattr(node, "peers", {}).items())
        else:
            peers = list(getattr(node, "peers", {}).items())

        for _peer_key, peer in peers:
            if not getattr(peer, "address", None):
                continue
            route_request = Message(
                topic=MessageTopic.ROUTE_REQUEST,
                content=b"\x01",
                sender=node.relay_public_key,
            )
            route_request.encrypt(peer.shared_key_bytes)
            enqueue_outgoing(
                node,
                peer.address,
                message=route_request,
                difficulty=peer.difficulty,
            )

        wait_deadline = time.time() + float(node.config.get("peer_timeout_interval", 10))
        while time.time() < wait_deadline:
            if getattr(validation_route, "peers", None):
                has_validators = True
                break
            time.sleep(0.1)

    if not has_validators:
        raise RuntimeError("no validator available")

    validators = {}
    for peer in validation_route.peers.values():
        if peer is not None and getattr(peer, "address", None):
            validators[getattr(peer, "public_key_bytes", None)] = peer
    for bucket in validation_route.buckets.values():
        for peer_key in bucket:
            peer = node.get_peer(peer_key)
            if peer is not None and getattr(peer, "address", None):
                validators[peer.public_key_bytes] = peer

    if not validators:
        raise RuntimeError("no validator available")

    for peer in validators.values():
        tx_message = Message(
            topic=MessageTopic.TRANSACTION,
            content=bytes(tx_hash),
            sender=node.relay_public_key,
        )
        tx_message.encrypt(peer.shared_key_bytes)
        enqueue_outgoing(
            node,
            peer.address,
            message=tx_message,
            difficulty=peer.difficulty,
        )

    return tx_hash



from __future__ import annotations

from typing import Any, Dict, Optional

from astreum.validation.models.fork import Fork


def load_forks(node: Any, data: Any) -> Dict[bytes, Fork]:
    """Load forks from fork dicts into node.forks."""
    for fork_data in data:
        fork = Fork.from_dict(fork_data)
        node.forks[fork.head] = fork

    return node.forks


def export_forks(node: Any) -> list[dict[str, Any]]:
    """Export node.forks as a list of fork dicts."""
    return [fork.to_dict() for fork in node.forks.values()]



from __future__ import annotations

import math
import time
from queue import Empty
from typing import Any, Callable

from ..models.account import Account
from ..models.accounts import Accounts
from ..models.block import Block
from ..models.transaction import Transaction, apply_transaction
from ..validator import current_validator
from ...storage.models.atom import Atom, AtomKind, ZERO32, bytes_list_to_atoms
from ...communication.handlers.object_response import OBJECT_FOUND_LIST_PAYLOAD
from ...storage.advertisments import advertise_atoms
from ...communication.models.message import Message, MessageTopic
from ...communication.models.ping import Ping
from ...communication.difficulty import message_difficulty
from ...communication.outgoing_queue import enqueue_outgoing

validator_advertisment_limit_seconds = 15 * 60


def _collect_block_ads(node: Any, block: Block) -> list[bytes]:
    heads: list[bytes] = []
    if block.atom_hash and block.atom_hash != ZERO32:
        heads.append(block.atom_hash)
    if block.body_hash and block.body_hash != ZERO32:
        body_list_atom = node.get_atom_from_local_storage(block.body_hash)
        if body_list_atom and body_list_atom.data and body_list_atom.data != ZERO32:
            heads.append(body_list_atom.data)
    return heads


def _collect_transaction_ads(node: Any, transactions: list[Transaction]) -> list[bytes]:
    ads: list[bytes] = []
    for tx in transactions:
        if not tx.hash:
            continue
        ads.append(tx.hash)
        tx_chain = node.get_atom_list(tx.hash)
        if tx_chain:
            body_list_atom: Atom = tx_chain[-1]
            if (
                body_list_atom.kind is AtomKind.LIST
                and body_list_atom.data
                and body_list_atom.data != ZERO32
            ):
                ads.append(body_list_atom.data)
    return ads


def _collect_receipt_ads(receipt_ids: list[bytes]) -> list[bytes]:
    return [rid for rid in receipt_ids if rid and rid != ZERO32]


def _collect_account_ads(accounts_hash: bytes | None, account_atoms: list[Atom]) -> list[bytes]:
    ads: list[bytes] = []
    if accounts_hash and accounts_hash != ZERO32:
        ads.append(accounts_hash)
    for atom in account_atoms:
        if atom.kind is AtomKind.SYMBOL and atom.data == b"trie":
            ads.append(atom.object_id())
    return ads


def make_validation_worker(
    node: Any,
) -> Callable[[], None]:
    """Build the validation worker bound to the given node."""

    def _validation_worker() -> None:
        node.logger.info("Validation worker started")
        stop = node._validation_stop_event

        def _award_validator_reward(block: Block, reward_amount: int) -> None:
            """Credit the validator account with the provided reward."""
            if reward_amount <= 0:
                return
            accounts = getattr(block, "accounts", None)
            validator_key = getattr(block, "validator_public_key_bytes", None)
            if accounts is None or not validator_key:
                node.logger.debug(
                    "Skipping validator reward; accounts snapshot or key missing"
                )
                return
            try:
                validator_account = accounts.get_account(
                    address=validator_key, node=node
                )
            except Exception:
                node.logger.exception("Unable to load validator account for reward")
                return
            if validator_account is None:
                validator_account = Account.create()
            validator_account.balance += reward_amount
            accounts.set_account(validator_key, validator_account)

        while not stop.is_set():
            validation_public_key = getattr(node, "validation_public_key", None)
            if not validation_public_key:
                node.logger.debug("Validation public key unavailable; sleeping")
                time.sleep(0.5)
                continue

            latest_block_hash = getattr(node, "latest_block_hash", None)
            if not isinstance(latest_block_hash, (bytes, bytearray)):
                node.logger.warning("Missing latest_block_hash; retrying")
                time.sleep(0.5)
                continue

            node.logger.debug(
                "Querying current validator for block %s",
                latest_block_hash.hex()
                if isinstance(latest_block_hash, (bytes, bytearray))
                else latest_block_hash,
            )
            try:
                scheduled_validator, _ = current_validator(node, latest_block_hash)
            except Exception as exc:
                node.logger.exception("Unable to determine current validator: %s", exc)
                time.sleep(0.5)
                continue

            if scheduled_validator != validation_public_key:
                expected_hex = (
                    scheduled_validator.hex()
                    if isinstance(scheduled_validator, (bytes, bytearray))
                    else scheduled_validator
                )
                node.logger.debug("Current validator mismatch; expected %s", expected_hex)
                time.sleep(0.5)
                continue

            try:
                previous_block = Block.from_atom(node, latest_block_hash)
            except Exception:
                node.logger.exception("Unable to load previous block for validation")
                time.sleep(0.5)
                continue

            try:
                current_hash = node._validation_transaction_queue.get_nowait()
                queue_empty = False
            except Empty:
                current_hash = None
                queue_empty = True
                node.logger.debug(
                    "No pending validation transactions; generating empty block"
                )

            try:
                accounts_snapshot = Accounts(root_hash=previous_block.accounts_hash)
            except Exception:
                accounts_snapshot = None
                node.logger.warning("Unable to initialise accounts snapshot for block")

            new_block = Block(
                chain_id=getattr(node, "chain", 0),
                previous_block_hash=latest_block_hash,
                previous_block=previous_block,
                number=(previous_block.number or 0) + 1,
                timestamp=None,
                accounts_hash=previous_block.accounts_hash,
                transactions_total_fees=0,
                transactions_hash=None,
                receipts_hash=None,
                delay_difficulty=None,
                validator_public_key_bytes=validation_public_key,
                nonce=0,
                signature=None,
                accounts=accounts_snapshot,
                transactions=[],
                receipts=[],
            )
            node.logger.debug(
                "Creating block #%s extending %s",
                new_block.number,
                (
                    node.latest_block_hash.hex()
                    if isinstance(node.latest_block_hash, (bytes, bytearray))
                    else node.latest_block_hash
                ),
            )

            # we may want to add a timer to process part of the txs only on a slow computer
            total_fees = 0
            while current_hash is not None:
                try:
                    total_fees += apply_transaction(node, new_block, current_hash)
                except NotImplementedError:
                    tx_hex = (
                        current_hash.hex()
                        if isinstance(current_hash, (bytes, bytearray))
                        else current_hash
                    )
                    node.logger.warning("Transaction %s unsupported; re-queued", tx_hex)
                    node._validation_transaction_queue.put(current_hash)
                    time.sleep(0.5)
                    break
                except Exception:
                    tx_hex = (
                        current_hash.hex()
                        if isinstance(current_hash, (bytes, bytearray))
                        else current_hash
                    )
                    node.logger.exception("Failed applying transaction %s", tx_hex)

                try:
                    current_hash = node._validation_transaction_queue.get_nowait()
                except Empty:
                    current_hash = None

            new_block.transactions_total_fees = total_fees
            reward_amount = total_fees if total_fees > 0 else 1
            if total_fees == 0 and queue_empty:
                node.logger.debug("Awarding base validator reward of 1 aster")
            elif total_fees > 0:
                node.logger.debug(
                    "Collected %d aster in transaction fees for this block", total_fees
                )
            _award_validator_reward(new_block, reward_amount)

            # create an atom list of transactions, save the list head hash as the block's transactions_hash
            transactions = new_block.transactions or []
            tx_hashes = [bytes(tx.hash) for tx in transactions if tx.hash]
            head_hash, _ = bytes_list_to_atoms(tx_hashes)
            new_block.transactions_hash = head_hash
            node.logger.debug("Block includes %d transactions", len(transactions))
            transaction_atoms = []
            for tx in transactions:
                if not tx.hash:
                    continue
                atoms = Transaction.get_atoms(node, tx.hash)
                if atoms is None:
                    node.logger.debug(
                        "Unable to load transaction atoms for %s",
                        tx.hash.hex(),
                    )
                    continue
                transaction_atoms.extend(atoms)

            receipts = new_block.receipts or []
            receipt_atoms = []
            receipt_hashes = []
            for rcpt in receipts:
                receipt_id, atoms = rcpt.to_atom()
                receipt_atoms.extend(atoms)
                receipt_hashes.append(bytes(receipt_id))
            receipts_head, _ = bytes_list_to_atoms(receipt_hashes)
            new_block.receipts_hash = receipts_head
            node.logger.debug("Block includes %d receipts", len(receipts))

            account_atoms = []
            if new_block.accounts is not None:
                try:
                    account_atoms = new_block.accounts.update_trie(node)
                    new_block.accounts_hash = new_block.accounts.root_hash
                    node.logger.debug(
                        "Updated trie for %d cached accounts",
                        len(new_block.accounts._cache),
                    )
                except Exception:
                    node.logger.exception("Failed to update accounts trie for block")

            now = time.time()
            min_allowed = new_block.previous_block.timestamp + 1
            nonce_time_seconds = node.nonce_time_ms / 1000.0
            expected_blocktime = now + nonce_time_seconds
            new_block.timestamp = max(int(math.ceil(expected_blocktime)), min_allowed)

            new_block.delay_difficulty = Block.calculate_delay_difficulty(
                previous_timestamp=previous_block.timestamp,
                current_timestamp=new_block.timestamp,
                previous_difficulty=previous_block.delay_difficulty,
            )
            
            try:
                nonce_started = time.perf_counter()
                new_block.generate_nonce(difficulty=previous_block.delay_difficulty)
                elapsed_ms = int((time.perf_counter() - nonce_started) * 1000)
                setattr(node, "nonce_time_ms", elapsed_ms)
                node.logger.debug(
                    "Found nonce %s for block #%s at difficulty %s",
                    new_block.nonce,
                    new_block.number,
                    new_block.delay_difficulty,
                )
            except Exception:
                node.logger.exception("Failed while searching for block nonce")
                time.sleep(0.5)
                continue
            
            # wait until the block timestamp is reached before propagating
            now = time.time()
            if now > (new_block.timestamp + 2):
                node.logger.warning(
                    "Skipping block #%s propagation; timestamp %s already elapsed (now=%s)",
                    new_block.number,
                    new_block.timestamp,
                    now,
                )
                continue

            spread_delay = new_block.timestamp - now
            if spread_delay > 0:
                node.logger.debug(
                    "Delaying distribution for %.3fs to reach block timestamp %s",
                    spread_delay,
                    new_block.timestamp,
                )
                time.sleep(spread_delay)
                
            # atomize block
            new_block_hash, new_block_atoms = new_block.to_atom()
            
            # hot set block atoms
            for block_atom in new_block_atoms:
                atom_id = block_atom.object_id()
                node._hot_storage_set(atom_id, block_atom)

            # hot set receipt atoms
            for receipt_atom in receipt_atoms:
                atom_id = receipt_atom.object_id()
                node._hot_storage_set(atom_id, receipt_atom)

            # hot set transaction atoms
            for transaction_atom in transaction_atoms:
                atom_id = transaction_atom.object_id()
                node._hot_storage_set(atom_id, transaction_atom)

            # hot set account atoms
            for account_atom in account_atoms:
                atom_id = account_atom.object_id()
                node._hot_storage_set(atom_id, account_atom)

            expires_at = time.time() + validator_advertisment_limit_seconds
            advertisement_ids = []
            advertisement_ids.extend(_collect_block_ads(node, new_block))
            advertisement_ids.extend(_collect_transaction_ads(node, transactions))
            advertisement_ids.extend(_collect_receipt_ads(receipt_hashes))
            advertisement_ids.extend(_collect_account_ads(new_block.accounts_hash, account_atoms))
            if advertisement_ids:
                entries = [
                    (atom_id, OBJECT_FOUND_LIST_PAYLOAD, expires_at)
                    for atom_id in advertisement_ids
                ]
                node.add_atom_advertisements(entries)
                advertise_atoms(node, entries=entries)
            # put as own latest block hash
            node.latest_block_hash = new_block_hash
            node.latest_block = new_block
            node.logger.info(
                "Created block #%s with hash %s (%d atoms)",
                new_block.number,
                new_block_hash.hex(),
                len(new_block_atoms),
            )
            

            # ping all peers to update their records
            if node.outgoing_queue and node.peers:
                try:
                    with node.peers_lock:
                        peers = list(node.peers.items())
                except Exception:
                    peers = list(getattr(node, "peers", {}).items())

                if peers:
                    ping_payload = Ping(
                        is_validator=True,
                        difficulty=message_difficulty(node),
                        latest_block=new_block_hash,
                    ).to_bytes()

                    for peer_key, peer in peers:
                        peer_hex = (
                            peer_key.hex()
                            if isinstance(peer_key, (bytes, bytearray))
                            else peer_key
                        )
                        address = getattr(peer, "address", None)
                        if not address:
                            node.logger.debug(
                                "Skipping validator ping to %s; address missing",
                                peer_hex,
                            )
                            continue
                        try:
                            ping_msg = Message(
                                topic=MessageTopic.PING,
                                content=ping_payload,
                                sender=node.relay_public_key,
                            )
                            ping_msg.encrypt(peer.shared_key_bytes)
                            if enqueue_outgoing(
                                node,
                                address,
                                message=ping_msg,
                                difficulty=peer.difficulty,
                            ):
                                node.logger.debug(
                                    "Queued validator ping to %s (%s)",
                                    address,
                                    peer_key.hex()
                                    if isinstance(peer_key, (bytes, bytearray))
                                    else peer_key,
                                )
                            else:
                                node.logger.debug(
                                    "Dropped validator ping to %s (%s); outgoing queue full",
                                    address,
                                    peer_key.hex()
                                    if isinstance(peer_key, (bytes, bytearray))
                                    else peer_key,
                                )
                        except Exception:
                            node.logger.exception("Failed queueing validator ping to %s", address)

            # upload block atoms
            for block_atom in new_block_atoms:
                atom_id = block_atom.object_id()
                node._cold_storage_set(atom_id, block_atom)

            # upload receipt atoms
            for receipt_atom in receipt_atoms:
                atom_id = receipt_atom.object_id()
                node._cold_storage_set(atom_id, receipt_atom)

            # upload transaction atoms
            for transaction_atom in transaction_atoms:
                atom_id = transaction_atom.object_id()
                node._cold_storage_set(atom_id, transaction_atom)

            # upload account atoms
            for account_atom in account_atoms:
                atom_id = account_atom.object_id()
                node._cold_storage_set(atom_id, account_atom)

        node.logger.info("Validation worker stopped")

    return _validation_worker



"""Helpers related to disconnecting communication components."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from astreum.node import Node


_SOCKET_ATTRS: tuple[str, ...] = ("incoming_socket", "outgoing_socket")
_THREAD_ATTRS: tuple[str, ...] = (
    "incoming_populate_thread",
    "incoming_process_thread",
    "outgoing_thread",
    "peer_manager_thread",
    "latest_block_discovery_thread",
    "verify_thread",
    "consensus_validation_thread",
)


def _set_event(node: "Node", attr_name: str) -> None:
    event = getattr(node, attr_name, None)
    if event is not None:
        event.set()


def _close_socket(node: "Node", attr_name: str) -> None:
    sock = getattr(node, attr_name, None)
    if sock is None:
        return
    try:
        sock.close()
    except Exception as exc:  # pragma: no cover - defensive logging path
        node.logger.warning("Error closing %s: %s", attr_name, exc)


def disconnect_node(node: "Node") -> None:
    """Gracefully stop worker threads and close communication sockets."""
    node.logger.info("Disconnecting Astreum Node")

    _set_event(node, "communication_stop_event")
    _set_event(node, "_validation_stop_event")
    _set_event(node, "_verify_stop_event")

    for sock_name in _SOCKET_ATTRS:
        _close_socket(node, sock_name)

    for thread_name in _THREAD_ATTRS:
        thread = getattr(node, thread_name, None)
        if thread is None or not thread.is_alive():
            continue
        thread.join(timeout=1.0)

    node.is_connected = False
    node.logger.info("Node disconnected")
